import numpy as np
import numpy.random as random
import io


class ptr(int):
    def __call__(self, MEM):
        return MEM[self]


class Node(object):
    def __init__(self):
        # tree-node basic
        self.parent = ptr(0)
        self.just_happened = np.empty(shape=[2], dtype=int)
        self.child = np.empty(shape=[0, 0], dtype=ptr)

        # InformationSet emulated
        self.actA = np.empty(shape=[0], dtype=int)
        self.actB = np.empty(shape=[0], dtype=int)
        # self.x = np.empty(shape=[0], dtype=float)
        # self.y = np.empty(shape=[0], dtype=float)
        self.u = 0.0
        self.v = 0.0

        # logical pointer mechanism
        self.__graph__ = None
        self.__addr__ = ptr(0)

    def __repr__(self):
        buff = io.StringIO()
        print("{", file=buff)
        print("h:", self.h(), file=buff)
        print("actA:", self.actA, file=buff)
        print("actB:", self.actB, file=buff)
        # print("x:", self.x, file=buff)
        # print("y:", self.y, file=buff)
        print("u:", self.u, file=buff)
        print("v:", self.v, file=buff)
        print("}", end="", file=buff)
        return buff.getvalue()

    def __str__(self):
        return self.__repr__()

    def get_parent(self):
        parent: Node = self.__graph__.MEM[self.parent]
        return parent

    def h(self):
        n = self
        G = self.__graph__
        y = np.empty([0, 2], dtype=int)
        while n.__addr__ != 0:  # root address is 0
            y = np.append(y, n.just_happened[None, :], axis=0)
            n = G.MEM[n.parent]
        y = np.flip(y, axis=0)
        y = y.tolist()
        return y

    def get_child(self, i, j):
        child = self.__graph__.MEM[self.child[i, j]]
        return child

    def __call__(self, ij_seq):
        n = self
        for ij in ij_seq:
            i, j = ij
            idx = np.where(n.actA == i)
            jdx = np.where(n.actB == j)
            if len(idx[0]) and len(jdx[0]):
                n = n.get_child(idx[0][0], jdx[0][0])
            else:
                n = self.__graph__.MEM[0]
        return n

    def __pos__(self):  # like &self in C++
        return self.__addr__


class Tree(object):  # node manager
    def __init__(self):
        self.MEM = {}
        self.N = 16

    def new_node(self):
        # addr = len(self.MEM)
        if len(self.MEM) == 0:
            addr = ptr(0)
        else:
            if len(self.MEM) >= 0.75 * self.N:
                self.N *= 2
            while True:
                addr = random.randint(1, self.N)
                if addr not in self.MEM.keys():
                    break
        addr = ptr(addr)
        self.MEM[addr] = Node()
        self.MEM[addr].__graph__ = self
        self.MEM[addr].ptr = addr
        return self.MEM[addr]


def main():
    T = Tree()
    root = T.new_node()
    root.branch = np.array([0, 0])
    root.actA = np.array([1, 2, 3])
    root.actB = np.array([1, 2, 3])
    root.child = np.resize(root.child, [3, 3])
    # root.x = np.resize(root.x, [3])
    # root.y = np.resize(root.y, [3])

    for i in range(len(root.actA)):
        for j in range(len(root.actB)):
            n = T.new_node()
            root.child[i, j] = +n
            n.parent = +root
            n.branch = np.array([root.actA[i], root.actB[j]])

    print("root:", root)
    print("+root:", +root)
    print("root.child:")
    print(root.child)
    print("all children of root:")
    for i in range(len(root.actA)):
        for j in range(len(root.actB)):
            print(root.get_child(i, j).h())


if __name__ == "__main__":
    main()
